"""
========================================
  Introduction to Computer Programming
========================================

----------------------------------------------------
  lab 7 examples: timing and tuples
----------------------------------------------------
"""


from random import randrange
from time import perf_counter_ns

from simple_graphics import clear, paint, COLUMNS, ROWS


def time_for_a_million():
    t0 = perf_counter_ns()
    for _ in range(10**6):
        pass
    td = perf_counter_ns() - t0
    print(f'for loop to a million took {td} nanoseconds')


def mini_split(s):
    """
    Returns pair of strings based on s, split across the first occurrence of
    the space character (' ').

    >>> mini_split('Computer science rocks!')
    ('Computer', 'science rocks!')
    """
    i = s.find(' ')
    p = (s[:i], s[i+1:])
    return p


# our now familiar Ulam function, using divmod and pattern matching
def ulam(m):
    (q, r) = divmod(m, 2)
    if r == 0:
        n = q
    else:
        n = 3*m + 1
    return n


# timing a (somewhat useful) loop
def print_time_count_uppers(s):
    count = 0
    t0 = perf_counter_ns()
    for c in s:
        if c.is_upper():
            count += 1
    td = perf_counter_ns - t0
    print(f'found {count} uppercase letters in {td} nanoseconds')


# same idea, but returning result and elapsed time as a pair
def time_count_uppers(s):
    count = 0
    t0 = perf_counter_ns()
    for c in s:
        if c.is_upper():
            count += 1
    td = perf_counter_ns - t0
    return (count, td)


# this is very much like Python's built-in join method for strings
def join(xs, sep):
    """
    Returns string formed by gluing together strings inside tuple xs,
    with string sep between each element.

    >>> join(('alpha','beta','gamma'), ':')
    'alpha:beta:gamma'
    """
    if len(xs) > 0:
        s = xs[0]
        i = 1
        while i < len(xs):
            s = s + sep + xs[i]
            i += 1
    else:
        s = ''
    return s


# generate tuple of n random integers between start and stop
def make_random_tuple(n, start, stop):
    ns = ()
    for _ in range(n):
        r = randrange(start, stop)
        ns += (r,)
    return ns


# generate tuple of n pairs from (0, 0) through (n-1, n-1)
def make_diagonal(n):
    """
    >>> make_diagonal(5)
    ((0, 0), (1, 1), (2, 2), (3, 3), (4, 4))
    """
    ns = ()
    for i in range(n):
        ns += ((i, i),)
    return ns


# assumes points is a tuple of pairs (themselves tuples of length two) of
# integers and that the coordinates in each pair correspond to valid locations
# in a simple_graphics window.
# try, for example:
#
# >>> draw_points(make_diagonal(ROWS))
#
def draw_points(points):
    clear()
    for p in points:
        paint(p[0], p[1])


# return new tuple like xs but only with elements that satisfy is_ok
# is_ok is a function that returns a Boolean
def filter_tuple(xs, is_ok):
    ys = ()
    for x in xs:
        if is_ok(x):
            ys += (x,)
    return ys


# return tuple based on evaluating function f on integers
# from start (inclusive) to stop (exclusive)
def build_tuple(f, start, stop):
    ys = ()
    for x in range(start, stop):
        ys += (f(x),)
    return ys


# return new tuple like xs, but with function f applied to each element
def map_tuple(f, xs):
    ys = ()
    for x in xs:
        ys += (f(x),)
    return ys


# familiar function, but for tuples
# returns pair representing position and value where
# value is first occurrence of smallest value in tuple xs
# assume xs is not empty
def find_min(xs):
    min_pos = 0
    min_val = xs[0]
    for i in range(1, len(xs)):
        if xs[i] < min_val:
            min_pos = i
            min_val = xs[i]
    return (min_pos, min_val)


# returns pair with first part, tuple like xs but with smallest value removed
# and second part is the removed value
# uses find_min and tuple slicing
def split_on_min(xs):
    (min_pos, min_val) = find_min(xs)
    cut = xs[:min_pos] + xs[min_pos+1:]
    return (cut, min_val)


# familiar function, but for tuples
# returns new tuple, like xs, but in increasing order
# uses split_on_min
def selection_sort(xs):
    ys = ()
    zs = xs
    for _ in range(len(xs)):
        (zs, smallest) = split_on_min(zs)
        ys += (smallest,)
    return ys
